# Class to control seven-segment LED displays based on SC16739 chips
# Copyright 2022 Любезный В. В.
# This script is licensed under Mozilla public license, version 2.0
# https://www.mozilla.org/en-US/MPL/2.0/
# Version 1.0
# Last change: April 3, 2022

from typing import Dict, List
import RPi.GPIO as GPIO

class SC16739_7segment_display:

    FONT = {
        ' ': '',
        '0': 'ABCDEF',
        '1': 'BC',
        '2': 'ABDEG',
        '3': 'ABCDG',
        '4': 'BCFG',
        '5': 'ACDFG',
        '6': 'ACDEFG',
        '7': 'ABC',
        '8': 'ABCDEFG',
        '9': 'ABCDFG',
        'A': 'ABCEFG',
        'b': 'CDEFG',
        'C': 'ADEF',
        'c': 'DEG',
        'd': 'BCDEG',
        'E': 'ADEFG',
        'e': 'ABDEFG',
        'F': 'AEFG',
        'H': 'BCEFG',
        'h': 'CEFG',
        'I': 'BC',
        'J': 'BCD',
        'L': 'DEF',
        'n': 'CEG',
        'o': 'CDEG',
        'O': 'ABCDEF',
        'P': 'ABEFG',
        'r': 'EG',
        'U': 'BCDEF',
        'u': 'CDE',
        '-': 'G',
        '_': 'D',
        # Cyrillic letters
        'А': 'ABCEFG',
        'Б': 'ACDEFG',
        'г': 'EG',
        'Г': 'AEF',
        'E': 'ADEFG',
        'З': 'ABCDG',
        'Н': 'BCEFG',
        'о': 'CDEG',
        'О': 'ABCDEF',
        'П': 'ABCEF',
        'п': 'CEG',
        'С': 'ADEF',
        'с': 'DEG',
        'у': 'BCDFG',
        'Ч': 'BCFG',
        'Ь': 'CDEFG'
    }
    """
    Defines which segments in the digit (except point)
    must be turned on when the character outputs.
    Segment letters definition:
        __A__
       |     |
       |F    |B
       |__G__|
       |     |
       |E    |C
       |__D__|
    """

    def __init__(self,
            gpio_pins: Dict,
            transmission_sequence: List,
            debug: bool = False
        ) -> None:
        """
        Constructor

        Keyword arguments:
        gpio_pins -- GPIO pin numbers. Dictionary with elements:

            MODE -- GPIO mode according to RPi.GPIO mode constants.
                If this element is not present, GPIO will not be
                initialized.
            OE -- number of the GPIO pin connected to the OE pins of
                SC16739 display chips (display on/off inverted
                signal). If this element isn't present, a display
                will turn off using all-zero data transmission and
                turn on using actual data transmission.
            DCLK (mandatory) -- number of the GPIO pin connected to
                the DCLK pins of the SC16739 display chips (clock
                signal)
            LATCH (mandatory) -- number of the GPIO pin connected to
                the LATCH pins of the SC16739 display chips
                (sends all the current data to the segments).
            SDI (mandatory) -- number of the GPIO pin connedted to
                the SDI pin of the first SC16739 display chip
                in the chain (data input).


        transmission_sequence -- sequence of segment data transmission.
            The list of strings.
            Each string contains a segment letter
            (A, B, C, D, E, F, G, P)
            and then a digit index starting at 0.
            Segment letter definition:
            __A__
           |     |
           |F    |B
           |__G__|
           |     |
           |E    |C
           |__D__| .P

        debug -- debug flag. True value used to print debug information.
            Default value is False.

        """
        # Remember creating params
        self.gpio_pins = gpio_pins
        self.transmission_sequence = transmission_sequence
        self.debug = debug

        self.digits_count = len(self.transmission_sequence) >> 3
        """
        Count of digits in the display.
        It's count of segments divided by 8
        """

        self.segments_state = []
        """
        Current state of all the segments in the display.
        Default value of every segment is False.
        """
        for segment in self.transmission_sequence:
            self.segments_state.append(False)

        # Init GPIO, set GPIO pin modes and initial values
        GPIO.setwarnings(False)

        # MODE: non-mandatory element. If present, set GPIO mode
        if 'MODE' in self.gpio_pins:
            GPIO.setmode(self.gpio_pins.get('MODE'))

        # OE: non-mandatory element. If present, set pin to output mode
        # and send 1 (inactive level)
        if 'OE' in self.gpio_pins:
            GPIO.setup(self.gpio_pins.get('OE'), GPIO.OUT)
            GPIO.output(self.gpio_pins.get('OE'), 1)

        # DCLK: mandatory element. Set pin to output mode and send 0.
        GPIO.setup(self.gpio_pins.get('DCLK'), GPIO.OUT)
        GPIO.output(self.gpio_pins.get('DCLK'), 0)

        # LATCH: mandatory element. The same as DCLK
        GPIO.setup(self.gpio_pins.get('LATCH'), GPIO.OUT)
        GPIO.output(self.gpio_pins.get('LATCH'), 0)

        # SDI - mandatory element.
        GPIO.setup(self.gpio_pins.get('SDI'), GPIO.OUT)
        GPIO.output(self.gpio_pins.get('SDI'), 0)

        # Turn display off
        self.on_off(False)

    def send_output(self, current_states: bool) -> None:
        """
        Send current output to the display.
        Keyword arguments:
        current_states -- if True, the method will send current
            segment state values. Otherwise, zeros will be sent.
        """
        for segment in self.segments_state:
            # send current segment state or 0 to the SDI output
            GPIO.output(
                self.gpio_pins.get('SDI'),
                segment if current_states else False
            )
            # After value setting, we need to tick the DCLK output
            GPIO.output(self.gpio_pins.get('DCLK'), True)
            GPIO.output(self.gpio_pins.get('DCLK'), False)

        # After all the data are sent, tick LATCH
        GPIO.output(self.gpio_pins.get('LATCH'), True)
        GPIO.output(self.gpio_pins.get('LATCH'), False)

    def set_segment_state(self, segment: str, state: bool) -> None:
        """
        Set state of given segment to on or off

        Keyword arguments:
        segment -- segment string, as described in transmission_sequence
            argument of the constructor
        state -- True (if segment needs to be turned on) or False (off)
        """
        self.segments_state[self.transmission_sequence.index(segment)] = state

    def on_off(self, turn_on: bool) -> None:
        """
        Turn display on or off

        Keyword arguments:
        turn_on -- turn display on (if True) or off
        """
        if (self.debug):
            print('SC16739_7segment_display.on_off is called.')
            print('turn_on is', turn_on)
        if 'OE' in self.gpio_pins:
            # We have OE line, so we can just set its state
            GPIO.output(self.gpio_pins.get('OE'), not turn_on)
        else:
            # No OE line. Send zeros or the current output to the display segments
            self.send_output(turn_on)



    def set_char(self, digit_index: int, character: str) -> None:
        """
        Set character to the digit
        Keyword arguments:
        digit_index -- index of the digit (starting from 0)
        character -- character to put into the digit.
        """
        # If digit index is out of range, generate error
        if (digit_index < 0) | (digit_index >= self.digits_count):
            raise Exception('Digit index is out of range')
        # If character is not present in the font, generate error
        if character not in self.FONT:
            raise Exception('No character "' + character + '" in the font')
        # Set the state of the segments according to the font
        for letter in ['A', 'B', 'C', 'D', 'E', 'F', 'G']:
            self.set_segment_state(
                str(letter) + str(digit_index),
                (letter in self.FONT[character]))


    def display_string(self, string: str) -> None:
        """
        Display string at the display

        Keyword arguments:
        string -- string to display.
        """
        if (self.debug):
            print('SC16739_7segment_display.display_string is called')
            print('String is ', string)
        # Make a digit counter in the i variable
        i = 0
        # Set characters of the string to the appropriate digits
        for character in string:
            self.set_char(i, character)
            i = i + 1
        # Set remaining digits as spaces
        while (i < self.digits_count):
            self.set_char(i, ' ')
            i = i + 1
        # Send current output to the display
        self.send_output(True)
        # Turn display on, if OE signal is present
        if 'OE' in self.gpio_pins:
            self.on_off(True)

    def set_point(self, digit_index: int, state: bool):
        """
        Turn the point on or off at the defined digit

        Keyword arguments:
        digit_index -- index of the digit (starting from 0)
        state -- turn point on (if True) or off
        """
        if (self.debug):
            print('SC16739_7segment_display.set_point is called')
            print('Digit index is ', digit_index, ', state is ', state)
        # Set the state of the point segment in appropriate digit
        self.set_segment_state('P' + str(digit_index), state)
        # Send display output
        self.send_output(True)
        # Turn display on, if OE signal is used
        if 'OE' in self.gpio_pins:
            self.on_off(True)
